/*
**-----------------------------------------------------------------------------
**  tsld - temperature sensor logger daemon
**  Copyright (c) 2004  John Clarke <johnc+tsld@kirriwa.net>
**-----------------------------------------------------------------------------
**  $Id: tsld.c,v 1.15 2005/01/17 11:44:08 johnc Exp $
**-----------------------------------------------------------------------------
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**  
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**  
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
**-----------------------------------------------------------------------------
*/

static const char CvsId[] = "$Id: tsld.c,v 1.15 2005/01/17 11:44:08 johnc Exp $";

/*----------------------------------------------------------------------------*/

#include "config.h"
#include "tsld.h"
#include "utils.h"

/*----------------------------------------------------------------------------*/

static const char Version[] = "tsld.c v" VERSION;

/*----------------------------------------------------------------------------*/

#include <getopt.h>

/* getopt variables */
char *optarg;
int optind, opterr, optopt;

/*----------------------------------------------------------------------------*/

/* command line options */
static int Average = TRUE;                      /* -a */
static char const * Device = "/dev/ttyS0";      /* -s */
static int SamplePeriod = 60;                   /* -p */
#ifdef HAVE_RRDTOOL
    static char const * RrdFilename = NULL;     /* -R */
    static int NumSensors = MAX_SENSORS;        /* -n */
#endif
static char const * PidFilename = NULL;         /* -r */
static uid_t User;                              /* -u */
static gid_t Group;                             /* -g */

#ifdef HAVE_RRDTOOL
    static int RrdArgc = 3;
    static char RrdData[1024];
    static const char *RrdArgv[] = 
    {
        "update",
        NULL,   /* RrdFilename */
        RrdData,
        NULL
    };
#endif
/*----------------------------------------------------------------------------*/

/* print a simple usage message */
static void Usage(void)
{
    printf("tsld [-a] [-d] [-f log_filename] [-g group ] [-h] [-i id] [-l [log_directory]]\n"
           "     [-p period] "
           #ifdef HAVE_RRDTOOL
               "[-R file [-n num_sensors]] "
           #endif
           "[-r pid_file] [-s serial_device]"
           #ifdef HAVE_RRDTOOL
               "\n     "
           #endif
           "[-u user] [-v]\n");
}

/*----------------------------------------------------------------------------*/

static void ProcessArgs(int argc, char * const argv[])
{
    /* command line options */
    #ifdef HAVE_RRDTOOL
        static const char *ShortOptions = ":adf:g:hi:l:n:p:R:r:s:u:v";
    #else
        static const char *ShortOptions = ":adf:g:hi:l:p:r:s:u:v";
    #endif
    static const struct option LongOptions[] =
    {
        {"no-averaging", 0, 0, 'a'},
        {"debug", 0, 0, 'd'},
        {"log-file", 1, 0, 'f'}, {"file", 1, 0, 'f'},
        {"group", 1, 0, 'g'}, {"guid", 1, 0, 'g'},
        {"help", 0, 0, 'h'},
        {"id", 1, 0, 'i'},
        {"log-directory", 1, 0, 'l'}, {"directory", 1, 0, 'l'}, 
        {"sample-period", 1, 0, 'p'}, {"period", 1, 0, 'p'},
        #ifdef HAVE_RRDTOOL
            {"num-sensors", 1, 0, 'n'}, {"sensors", 1, 0, 'n'},
            {"rrd-database", 1, 0, 'R'},
        #endif
        {"pid-file", 1, 0, 'r'},
        {"device", 1, 0, 's'}, {"serial-device", 1, 0, 's'}, {"serial-port", 1, 0, 's'},
        {"uiser", 1, 0, 'u'}, {"uid", 1, 0, 'u'},
        {"version", 0, 0, 'v'}
    };
    int c;
    char *p;

    SerialPort = NULL;
    LogFile = NULL;
    LogDirectory = NULL;
    LogFilename = NULL;
    Id = NULL;
    Debug = 0;
    User = geteuid();
    Group = getegid();

    while (1)
    {
        c = getopt_long(argc, argv, ShortOptions, LongOptions, NULL);
        if (c == -1)
            break;
        
        switch (c)
        {
            case 'a':
                /* don't save average over sample period */
                if (Debug > 2)
                    printf("Got option: -a or --no-average\n");
                Average = FALSE;
                break;
                
            case 'd':
                /* increase debug level */
                if (Debug > 2)
                    printf("Got option: -d or --debug\n");
                Debug++;
                break;

            case 'f':
                /* set log filename */
                if (Debug > 2)
                    printf("Got option: -f or --log-file\n");
                if (LogDirectory != NULL)
                {
                    fprintf(stderr, "-f cannot be used with -l\n");
                    exit(EXIT_FAILURE);
                }
                if (LogFilename != NULL)
                {
                    fprintf(stderr, "-f cannot be used more than once\n");
                    exit(EXIT_FAILURE);
                }

                LogFilename = optarg;
                break;

            case 'g':
                /* set gid */
                if (Debug > 2)
                    printf("Got option: -g or --group\n");
                {
                    gid_t g;

                    g = strtol(optarg, &p, 10);
                    if (g == 0 && p == optarg)
                    {
                        /* not a number, try a name */
                        g = GetGid(optarg);
                        if (g == (gid_t)-1)
                        {
                            fprintf(stderr, "Group %s unknown\n", optarg);
                            exit(EXIT_FAILURE);
                        }
                    }
                    /* super user only */
                    /* FIX: allow users to change groups */
                    if (geteuid() != 0 && g != getegid())
                        fprintf(stderr, "-g ignored, only root is allowed to use it\n");
                    else Group = g;
                }
                break;
                
            case 'h':
                /* help */
                if (Debug > 2)
                    printf("Got option: -h or --help\n");
                Usage();
                exit(EXIT_SUCCESS);
                break;
                
            case 'i':
                /* set id */
                if (Debug > 2)
                    printf("Got option: -i or --id\n");
                Id = optarg;
                break;

            case 'l':
                /* set log directory */
                if (Debug > 2)
                    printf("Got option: -l or --log-directory\n");
                if (LogFilename != NULL)
                {
                    fprintf(stderr, "-l cannot be used with -d\n");
                    exit(EXIT_FAILURE);
                }
                if (LogDirectory != NULL)
                {
                    fprintf(stderr, "-l cannot be used more than once\n");
                    exit(EXIT_FAILURE);
                }
                LogDirectory = optarg;
                /* check if directory exists and is writeable */
                if (!DirExists(LogDirectory))
                {
                    fprintf(stderr, "Log directory %s does not exist\n", LogDirectory);
                    exit(EXIT_FAILURE);
                }
                if (access(LogDirectory, W_OK) == -1)
                {
                    fprintf(stderr, "Log directory %s is not writeable: %s\n", LogDirectory, strerror(errno));
                    exit(EXIT_FAILURE);
                }
                break;

            case 'p':
                /* set sample period */
                if (Debug > 2)
                    printf("Got option: -p or --sample-period\n");
                SamplePeriod = atoi(optarg);
                if (SamplePeriod < MIN_SAMPLE_PERIOD)
                {
                    fprintf(stderr, "Sample period must be at least five seconds\n");
                    exit(EXIT_FAILURE);
                }
                break;

            #ifdef HAVE_RRDTOOL
                case 'n':
                    if (Debug > 2)
                        printf("Got option: -n or --num-sensors\n");
                    NumSensors = atoi(optarg);
                    if (NumSensors < 1 || NumSensors > MAX_SENSORS)
                    {
                        fprintf(stderr, "Number of sensors must be between 1 and %d\n", MAX_SENSORS);
                        exit(EXIT_FAILURE);
                    }
                    break;

                case 'R':
                    /* log temperatures with rrdtool */
                    if (Debug > 2)
                        printf("Got option: -R or --rrd-database\n");
                    RrdFilename = optarg;
                    RrdArgv[1] = RrdFilename;
                    if (access(RrdFilename, R_OK | W_OK) == -1)
                    {
                        fprintf(stderr, "Rrd database file %s needs to be readable and writeable: %s\n", RrdFilename, strerror(errno));
                        exit(EXIT_FAILURE);
                    }
                    break;
            #endif
                
            case 'r':
                /* save daemon pid in this file */
                if (Debug > 2)
                    printf("Got option: -r or --pid-file\n");
                PidFilename = optarg;
                break;
                
            case 's':
                /* set serial port device filename */
                if (Debug > 2)
                    printf("Got option: -s or --serial-device\n");
                Device = optarg;
                break;
                
            case 'u':
                /* set uid */
                if (Debug > 2)
                    printf("Got option: -u or --user\n");
                {
                    uid_t u;

                    u = strtol(optarg, &p, 10);
                    if (u == 0 && p == optarg)
                    {
                        u = GetUid(optarg);
                        if (u == (uid_t)-1)
                        {
                            fprintf(stderr, "User %s unknown\n", optarg);
                            exit(EXIT_FAILURE);
                        }
                    }
                    /* super user only */
                    /* FIX: default to user's primary group if -g not given */
                    if (geteuid() != 0 && u != geteuid())
                        fprintf(stderr, "-u ignored, only root is allowed to use it\n");
                    else User = u;
                }
                break;
                
            case 'v':
                /* display version number */
                if (Debug > 2)
                    printf("Got option: -v or --version\n");
                printf("%s\n", Version);
                exit(EXIT_SUCCESS);
                break;

            case ':':
                /* missing parameter */
                fprintf(stderr, "`-%c' needs a parameter.  Run `tsld -h' for help.\n", optopt);
                exit(EXIT_FAILURE);
                break;
                
            case '?':
                /* unknown option */
                fprintf(stderr, "Unrecognised option: `-%c'.  Run `tsld -h' for help.\n", optopt);
                exit(EXIT_FAILURE);
                break;
        }
    }

    if (Debug > 4)
        exit(EXIT_SUCCESS);

    if (LogDirectory == NULL
        #ifdef HAVE_RRDTOOL
            && RrdFilename == NULL
        #endif
        && LogFilename == NULL)
    {
        fprintf(stderr, "No logging options specified; exiting\n");
        exit(EXIT_FAILURE);
    }

    if (LogFilename != NULL)
    {
        if (*LogFilename == '/')
        {
            /* absolute pathname */
            if (Id != NULL)
            {
                p = LogFilename;
                LogFilename = Malloc(strlen(LogFilename) + strlen(Id) + 2);
                strcpy(LogFilename, p);
                strcat(LogFilename, "_");
                strcat(LogFilename, Id);
            }
        }
        else 
        {
            p = LogFilename;
            /* relative pathname, convert to absolute */
            LogFilename = Malloc(PATH_MAX + 1);
            if (LogFilename == NULL)
            {
                perror("malloc");
                exit(EXIT_FAILURE);
            }
            if (getcwd(LogFilename, PATH_MAX) != 0)
            {
                perror("getcwd");
                exit(EXIT_FAILURE);
            }
            if (p[strlen(p) - 1] != '/')
                strncat(LogFilename, "/", PATH_MAX - strlen(LogFilename));
            strncat(LogFilename, p, PATH_MAX - strlen(LogFilename));
            if (Id != NULL)
            {
                strncat(LogFilename, "_", PATH_MAX * 2 - strlen(LogFilename));
                strncat(LogFilename, Id, PATH_MAX * 2 - strlen(LogFilename));
            }
        }
    }
    if (Debug)
    {
        printf("Starting tsld on device %s, logging %ssamples every %d seconds",
               Device, (Average ? "averaged " : ""), SamplePeriod);
        if (LogFilename != NULL)
            printf(" into file %s", LogFilename);
        else if (LogDirectory != NULL)
            printf(" into directory %s", LogDirectory);
        printf("\n");
    }
}

/*----------------------------------------------------------------------------*/

static void ReadSensor(void)
{
    char *p;
    char Buffer[1024];                                      /* buffer to hold a line from the sensor */
    char Degree = ' ';                                      /* Celsius or Farenheit */
    int Sensor;                                             /* sensor number - the device supports up to four sensors */
    float Temp;
    float TotalTemp[MAX_SENSORS];                           /* the sum of samples for averaging */
    int NSamples[MAX_SENSORS];                              /* the number of samples for averaging */
    float LastTemp[MAX_SENSORS];                            /* last temperature read from each sensor */
    SensorState State[MAX_SENSORS];                         /* flags to indicate the state of each sensor */
    time_t LastSampleTime;                                  /* time last sample written to log */
    time_t Now;                                             /* current time */

    /* initialise sample variables */    
    Now = time(NULL);
    for (Sensor = 0; Sensor < MAX_SENSORS; Sensor++)
    {
        TotalTemp[Sensor] = 0.0;
        NSamples[Sensor] = 0;
        LastTemp[Sensor] = 0.0;
        State[Sensor] = NotPresent;
    }
    LastSampleTime = (Now / SamplePeriod) * SamplePeriod;

    /* loop forever, reading samples from the sensor and logging them */
    while (1)
    {
        ReadSerialPort(Buffer, sizeof(Buffer));
        Now = time(NULL);
        p = Buffer;
        while (*p == '\r' || *p == '\n')
            p++;
        if (*p == 'R')
        {
            /* reset packet */
            /* e.g. "R V1.0 2002-01-06 20:37:37 C" */
            if (Debug > 1)
                Message(stdout, LOG_INFO, "%s", p);
            if (sscanf(p, "R V%*f %*d-%*d-%*d %*d:%*d:%*d %c", &Degree) == 1)
            {
                LogTime(Now);
                fprintf(LogFile, " %s", p);
            }
            else
            {
                Message(stderr, LOG_WARNING, "Unrecognised reset packet: %s\n", p);
                Degree = ' ';
            }
        }
        else if (isdigit(*p) && sscanf(p, "%d %f", &Sensor, &Temp) == 2)
        {
            /* data packet */
            /* e.g. "1 0019.87" */

            Sensor--;
            
            State[Sensor] = Alive;
            LastTemp[Sensor] = Temp;

            if (Average)
            {
                TotalTemp[Sensor] += Temp;
                NSamples[Sensor]++;
            }
            else TotalTemp[Sensor] = Temp;

            if (Now >= LastSampleTime + SamplePeriod)
            {
                /* log sample */

                LastSampleTime += SamplePeriod;

                LogTime(Now);
                #ifdef HAVE_RRDTOOL
                    if (RrdFilename != NULL && Debug <= 3)
                        sprintf(RrdData, "%u", (unsigned)Now);
                #endif

                for (Sensor = 0; Sensor < MAX_SENSORS; Sensor++)
                {
                    switch (State[Sensor])
                    {
                        case NotPresent:
                            if (Debug <= 3)
                            {
                                fprintf(LogFile, " -");
                                #ifdef HAVE_RRDTOOL
                                    if (RrdFilename != NULL && Sensor < NumSensors)
                                        strcat(RrdData, ":U");
                                #endif
                            }
                            break;
                        case Silent:
                            if (Debug <= 3)
                            {
                                fprintf(LogFile, " *");
                                Message(stderr, LOG_WARNING, "Sensor %d has stopped reporting\n", Sensor + 1);
                                #ifdef HAVE_RRDTOOL
                                    if (RrdFilename != NULL && Sensor < NumSensors)
                                        strcat(RrdData, ":U");
                                #endif
                            }
                            break;
                        case Alive:
                            State[Sensor] = Silent;
                            Temp = (Average ? TotalTemp[Sensor] / (float)NSamples[Sensor]
                                            : TotalTemp[Sensor]);
                            if (Debug <= 3)
                            {
                                fprintf(LogFile, " %-5.2f", Temp);
                                #ifdef HAVE_RRDTOOL
                                    if (RrdFilename != NULL && Sensor < NumSensors)
                                        sprintf(&RrdData[strlen(RrdData)], ":%f", Temp);
                                #endif
                            }
                            break;
                    }
                    TotalTemp[Sensor] = 0.0;
                    NSamples[Sensor] = 0;
                }
                if (Debug <= 3)
                {
                    fprintf(LogFile, "\n");
                    #ifdef HAVE_RRDTOOL
                        if (RrdFilename != NULL)
                        {
                            /* 
                            **  rrd_update() uses getopt which crashes if 
                            **  optind isn't cleared.  To be safe, clear
                            **  all variables.
                            */
                            optarg = NULL;
                            optind = opterr = optopt = 0;
                            if (rrd_update(RrdArgc, (char **)&RrdArgv) != 0)
                                Message(stderr, LOG_ERR, "rrd_update failed: %s\n", rrd_get_error());
                        }
                    #endif
                }
            }
            if (Debug > 1)
                printf("%s", p);
            else if (Debug)
            {
                int AnyAlive = 0;

                for (Sensor = 0; Sensor < MAX_SENSORS; Sensor++)
                {
                    if (State[Sensor] == Alive)
                    {
                        if (AnyAlive)
                            printf("  |  ");
                        printf("% 5.2f%c", LastTemp[Sensor], Degree);
                        AnyAlive = 1;
                    }
                }
                if (AnyAlive)
                    printf("\r");
            }
        }
        else if (Debug)
        {
            Message(stderr, LOG_WARNING, "Unrecognised data  packet: %s\n", p);
        }

        if (SigUsr1Received)
        {
            SigUsr1Received = FALSE;
            
            Message(stdout, LOG_INFO, "Resetting sensor on SIGUSR1\n");
            ResetSensor();
        }

        if (SigHupReceived)
        {
            SigHupReceived = FALSE;
            
            Message(stdout, LOG_INFO, "Rotating logs on SIGHUP\n");
            ReopenLogFile();
        }
    }
}

/*----------------------------------------------------------------------------*/

int main(int argc, char * const argv[])
{
    ProcessArgs(argc, argv);

    /* fork daemon if not in debug mode */
    if (!Debug)
    {
        pid_t Pid;

        Pid = fork();
        if (Pid == -1)
        {
            /* failed */
            perror("fork");
            exit(EXIT_FAILURE);
        }
        else if (Pid != 0)
        {
            /* parent, exit */
            printf("Forked daemon pid %d\n", Pid);
            exit(EXIT_SUCCESS);
        }
        
        /* child */

        #if HAVE_SYSLOG
            /* log startup */
            openlog("tsld", LOG_PID | LOG_NOWAIT, LOG_DAEMON);
            syslog(LOG_DAEMON | LOG_NOTICE, "daemon started\n");
        #endif

        /* detach from the terminal */
        fclose(stdout);
        fclose(stderr);

        /* change to the root directory and become session leader */
        chdir("/");
        setsid();
    }
    
    /* install atexit handler to cleanup */
    atexit(CleanupAndExit);

    if (PidFilename != NULL)
    {
        /* save pid */
        FILE *PidFile;
        
        PidFile = Open(PidFilename, "w");
        if (PidFile != NULL)
        {
            fprintf(PidFile, "%d\n", getpid());
            fclose(PidFile);
        }
    }
    
    /* change to a non-privileged user/group */
    if (User == 0)
    {
        User = GetUid("nobody");
        if (User == (uid_t)-1)
        {
            Message(stderr, LOG_ERR, "Fatal error: UID 0 not allowed and user nobody doesn't exist\n");
            exit(EXIT_FAILURE);
        }
        Message(stderr, LOG_WARNING, "UID 0 not allowed; defaulting to nobody (%d)\n", User);
    }
    
    if (Group == 0)
    {
        Group = GetUid("nobody");
        if (Group == (gid_t)-1)
        {
            Message(stderr, LOG_ERR, "Fatal error: GID 0 not allowed and group nobody doesn't exist\n");
            exit(EXIT_FAILURE);
        }
        Message(stderr, LOG_WARNING, "GID 0 not allowed; defaulting to nobody (%d)\n", Group);
    }
    
    if (setgid(Group) != 0)
    {
        Message(stderr, LOG_ERR, "setgid: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }
    if (setuid(User) != 0)
    {
        Message(stderr, LOG_ERR, "setuid: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }

    SetupSignalHandlers();

    /* connect to the temperature sensor */
    OpenSerialPort(Device);

    ResetSensor();
    
    ReopenLogFile();

    /* do the real work */
    ReadSensor();
    
    if (Debug)
        Message(stdout, LOG_NOTICE, "Exiting\n");
    else Message(stdout, LOG_NOTICE, "daemon stopped\n");

    return EXIT_SUCCESS;
}
