/*
**-----------------------------------------------------------------------------
**  tsld - temperature sensor logger daemon
**  Copyright (c) 2004  John Clarke <johnc+tsld@kirriwa.net>
**-----------------------------------------------------------------------------
**  $Id: utils.c,v 1.1 2005/01/17 11:38:34 johnc Exp $
**-----------------------------------------------------------------------------
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**  
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**  
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
**-----------------------------------------------------------------------------
*/

#include "config.h"
#include "tsld.h"
#include "utils.h"

/*----------------------------------------------------------------------------*/

static const char CvsId[] = "$Id: utils.c,v 1.1 2005/01/17 11:38:34 johnc Exp $";

/*----------------------------------------------------------------------------*/

int SigHupReceived = FALSE;             /* flag to indicate HUP signal */
int SigUsr1Received = FALSE;            /* flag to indicate USR1 signal */

/*----------------------------------------------------------------------------*/

/* print a message either to the given stream or to syslog with the given level */
void Message(FILE *Stream, int Level, char const * Format, ...)
{
    va_list ap;

    va_start(ap, Format);
    if (Debug)
        vfprintf(Stream, Format, ap);
    #if HAVE_SYSLOG
        else
            vsyslog(LOG_DAEMON | Level, Format, ap);
    #endif
    va_end(ap);
}

/*----------------------------------------------------------------------------*/

/* called on exit to close files and restore serial port settings */
void CleanupAndExit(void)
{
    if (SerialPort != NULL)
    {
        /* restore serial port state */
        tcsetattr(fileno(SerialPort), TCSAFLUSH, &SavedTerminalAttributes);
        /* and close it */
        fclose(SerialPort);
        SerialPort = NULL;
    }
    if (LogFile != NULL)
    {
        /* and close the log file */
        fclose(LogFile);
        LogFile = NULL;
    }

    exit(EXIT_FAILURE);
}

/*----------------------------------------------------------------------------*/

/* wrapper around malloc with error checking */
void *Malloc(size_t Size)
{
    void *p;
    
    p = malloc(Size);
    if (p != NULL)
        return p;

    perror("malloc");
    exit(EXIT_FAILURE);
}

/*----------------------------------------------------------------------------*/

/* opens a file */
FILE *Open(char const * Filename, const char *Mode)
{
    FILE *F;

    if (Debug > 3)
        return NULL;

    F = fopen(Filename, Mode);
    if (F == NULL)
    {
        fprintf(stderr, "fopen: %s: %s\n", Filename, strerror(errno));
        exit(EXIT_FAILURE);
    }

    return F;
}

/*----------------------------------------------------------------------------*/

/* check if Dir exists */
int DirExists(char const *Dir)
{
    struct stat Stat;

    if (stat(Dir, &Stat) == -1)
    {
        /* try to create it if it doesn't exist */
        if (errno == ENOENT)
            return FALSE;
        Message(stderr, LOG_ERR, "Cannot stat %s: %s\n", Dir, strerror(errno));
        exit(EXIT_FAILURE);
    }
    else if (S_ISDIR(Stat.st_mode))
        return TRUE;

    Message(stderr, LOG_ERR, "%s exists and is not a directory\n", Dir);
    exit(EXIT_FAILURE);
}

/* create Dir if it doesn't exist */
static void CreateDirIfNotExist(char const *Dir)
{
    if (Debug > 3)
        return;

    if (DirExists(Dir) || mkdir(Dir, 0755) == 0)
        return;

    Message(stderr, LOG_ERR, "Cannot create log directory %s: %s\n", Dir, strerror(errno));
    exit(EXIT_FAILURE);
}

/*----------------------------------------------------------------------------*/

/* close and reopen the log file */
void ReopenLogFile(void)
{
    if (Debug > 3)
        return;

    if (LogFile != NULL)
    {
        if (Debug)
            Message(stdout, LOG_INFO, "Closing log file          \n");
        fclose(LogFile);
    }
    if (LogDirectory != NULL)
    {
        time_t Now = time(NULL);
        struct tm *LocalTime = localtime(&Now);

        if (LogFilename == NULL)
        {
            int Length = strlen(LogDirectory)
                         + 1                        /* '/' */
                         + 4                        /* year */
                         + 1                        /* '/' */
                         + 2                        /* month */
                         + 1                        /* '/' */
                         + 2                        /* day */
                         + 1;                       /* null terminator */
            
            if (Id != NULL)
                Length += 1                         /* '_' */
                          + strlen(Id);

            
            LogFilename = Malloc(strlen(LogDirectory)
                                 + 1    /* '/' */
                                 + 4    /* year */
                                 + 1    /* '/' */
                                 + 2    /* month */
                                 + 1    /* '/' */
                                 + 2    /* day */
                                 + 1);  /* null terminator */
        }

        sprintf(LogFilename, "%s/%04d", LogDirectory,
                LocalTime->tm_year + 1900);
        CreateDirIfNotExist(LogFilename);
        
        sprintf(LogFilename, "%s/%02d", LogFilename,
                LocalTime->tm_mon + 1);
        CreateDirIfNotExist(LogFilename);

        sprintf(LogFilename, "%s/%02d", LogFilename,
                LocalTime->tm_mday);
        if (Id != NULL)
        {
            strcat(LogFilename, "_");
            strcat(LogFilename, Id);
        }
        
        LogDay = LocalTime->tm_yday;
    }
    if (LogFilename != NULL)
    {
        if (Debug)
            Message(stdout, LOG_INFO, "Opening log file %s         \n", LogFilename);
        LogFile = Open(LogFilename, "a");
        if (LogFile == NULL)
            Message(stderr, LOG_ERR, "Cannot open log file %s: %s\n", LogFilename, strerror(errno));
        setvbuf(LogFile, NULL, _IOLBF, 0);
    }
}


void LogTime(const time_t Time)
{
    struct tm *LocalTime = localtime(&Time);
    
    if (Debug > 3)
        return;

    if (LogDirectory != NULL && LocalTime->tm_yday != LogDay)
        ReopenLogFile();

    fprintf(LogFile, "[%04d-%02d-%02d %02d:%02d:%02d] %u",
                     LocalTime->tm_year + 1900, 
                     LocalTime->tm_mon + 1,
                     LocalTime->tm_mday,
                     LocalTime->tm_hour,
                     LocalTime->tm_min,
                     LocalTime->tm_sec, 
                     (unsigned int)Time);
}

/*----------------------------------------------------------------------------*/

/* opens the serial port and configures it to talk to the sensor */
void OpenSerialPort(char const * Device)
{
    struct termios T;

    if (Debug > 3)
        return;

    /* open the serial port */
    SerialPort = fopen(Device, "r");
    if (SerialPort == NULL)
    {
        Message(stderr, LOG_ERR, "fopen: %s: %s\n", Device, strerror(errno));
        exit(EXIT_FAILURE);
    }

    /* save setup */
    tcgetattr(fileno(SerialPort), &SavedTerminalAttributes);
    T = SavedTerminalAttributes;

    /* set comms parameters */
    
    /* 2400 */
    if (cfsetispeed(&T, B2400) != 0)
    {
        Message(stderr, LOG_ERR, "cfsetispeed: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }
    if (cfsetospeed(&T, B2400) != 0)
    {
        Message(stderr, LOG_ERR, "cfsetospeed: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }
    /* 8 bits */
    T.c_cflag = (T.c_cflag & ~CSIZE ) | CS8;
    /* no parity */
    T.c_cflag &= ~PARENB;
    /* 1 stop bit */
    T.c_cflag &= ~CSTOPB;
    /* drop DTR on close */
    T.c_cflag |= HUPCL;

    if (tcsetattr(fileno(SerialPort), TCSAFLUSH, &T) != 0)
    {
        Message(stderr, LOG_ERR, "tcsetattr: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }

    if (Debug)
        Message(stdout, LOG_DEBUG, "Serial device %s opened: 2400/8/N/1\n", Device);
}

/*----------------------------------------------------------------------------*/

/* receive a line of text from the sensor */
void ReadSerialPort(char *Buffer, int BufferSize)
{
    if (Debug > 3)
    {
        sleep(1);
        strncpy(Buffer, "1 12.34\n", BufferSize);
        return;
    }

    if (fgets(Buffer, BufferSize, SerialPort) == NULL)
    {
        Message(stderr, LOG_ERR, "fgets: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }
}

/*----------------------------------------------------------------------------*/

/* reset the sensor by lowering and raising DTR */
void ResetSensor(void)
{
    int Signals;

    if (Debug > 3)
        return;

    ioctl(fileno(SerialPort), TIOCMGET, &Signals);

    Signals &= ~TIOCM_DTR;
    ioctl(fileno(SerialPort), TIOCMSET, &Signals);

    sleep(1);

    Signals |= TIOCM_DTR;
    ioctl(fileno(SerialPort), TIOCMSET, &Signals);
}

/*----------------------------------------------------------------------------*/

/* gets the uid associated with the given name */
uid_t GetUid(const char *Username)
{
    struct passwd *Passwd;

    Passwd = getpwnam(Username);
    if (Passwd == NULL)
        return -1;

    return Passwd->pw_uid;
}

/* gets the gid associated with the given name */
uid_t GetGid(const char *Groupname)
{
    struct group *Group;

    Group = getgrnam(Groupname);
    if (Group == NULL)
        return -1;

    return Group->gr_gid;
}

/*----------------------------------------------------------------------------*/

/* USR1 signal handler */
static void SigUsr1(int Signal)
{
    SigUsr1Received = 1;
}

/* HUP signal handler */
static void SigHup(int Signal)
{
    SigHupReceived = 1;
}

/* TERM signal handler */
static void SigTerm(int Signal)
{
    Message(stdout, LOG_INFO, "Exiting on signal %d\n", Signal);
    CleanupAndExit();
}

/* INT signal handler */
static void SigInt(int Signal)
{
    Message(stdout, LOG_INFO, "Exiting on signal %d\n", Signal);
    CleanupAndExit();
}

/* QUIT signal handler */
static void SigQuit(int Signal)
{
    Message(stdout, LOG_INFO, "Exiting on signal %d\n", Signal);
    CleanupAndExit();
}

static void InstallSignalHandler(int Signal, void (*Handler)(int))
{
    struct sigaction Sa;

    memset(&Sa, 0, sizeof(Sa));
    Sa.sa_handler = Handler;
    Sa.sa_flags = SA_NOCLDSTOP | SA_RESTART;
    if (sigaction(Signal, &Sa, NULL) != 0)
        Message(stderr, LOG_ERR, "Cannot install signal %d handler: %s\n", Signal, strerror(errno));
    if (Debug > 1)
        Message(stdout, LOG_DEBUG, "Signal %d handler installed\n", Signal);
}

/* installs the signal handlers */
void SetupSignalHandlers(void)
{
    /* setup signal handlers */
    InstallSignalHandler(SIGTERM, SigTerm);
    InstallSignalHandler(SIGHUP, SigHup);
    InstallSignalHandler(SIGUSR1, SigUsr1);
    if (Debug)
    {
        InstallSignalHandler(SIGINT, SigInt);
        InstallSignalHandler(SIGQUIT, SigQuit);
    }
}
